#!/usr/bin/env python:
#-*- coding: utf-8 -*-

########################################################################
# Fichier Polynome.py                                                  #
# (C) Alexandre Casamayou-Boucau, Pascal Chauvin, Guillaume Connan     #
#                                                                      #
# Complément de l'ouvrage :                                            #
# Programmation en Python pour les mathématiques                       #
# Editeur : Dunod                   -        Collection : Sciences Sup #
# ISBN-10: 2100574221               -          ISBN-13: 978-2100574223 #
#                                   -                                  #
# Version du 18/01/2012             -                  Licence : GPLv2 #
######################################################################## 

############################################################################
##
##           Chap 3 / § 1 :  M A T R I C E S
##
############################################################################


### préambule

def lignes(M):
    # return range(len(M)) # sur Python 2.7
     return [k for k in range(len(M))] # sur python3

def cols(M):
    # return range(len(M[0])) # sur Python 2.7
     return [k for k in range(len(M[0]))] # sur python3

    
### transposée

def transpose(M):
    return [[M[i][j] for i in lignes(M)] for j in cols(M)]

### somme de matrices

def msom(A,B):
    if lignes(A) != lignes(B) or cols(A) != cols(B):
        return "Tailles non compatibles"
    else:
        return [[A[i][j] + B[i][j] for j in cols(A)] for i in lignes(A)]



### produit par un scalaire

def mprod_scal(M,k):
    return [[k * M[i][j] for j in cols(M)] for i in lignes(M)]


### produit de matrices

def mprod1(A,B):
    if cols(A) != lignes(B):
        return "Problème de taille"
    else:
        C = [[0 for p in cols(B)] for p in lignes(A)]
        for i in lignes(A):
            for j in cols(B):
                for k in lignes(B):
                    C[i][j]+=A[i][k]*B[k][j]
    return C

# on crée une fonction qui calcule la somme
# des éléments d'une liste
def som_liste(L):
    s = 0
    for i in range(len(L)):
        s+=L[i]
    return s

def mprod2(A,B):
    if cols(A) != lignes(B):
        return "Problème de taille"
    else:
        return [[som_liste([A[i][k]*B[k][j] for k in lignes(B)]) for j in cols(B)] for i in lignes(A)]

def mprod3(A,B):
    if cols(A) != lignes(B):
        return "Problème de taille"
    else:
        C = [[0 for p in cols(B)] for p in lignes(A)]
        for i in lignes(A):
            Ai,Ci = A[i],C[i]
            for j in cols(B):
                for k in lignes(B):
                    Ci[j]+=Ai[k]*B[k][j]
    return C

def mprod(A,B):
    tB = transpose(B)
    if cols(A) != lignes(B):
        return "Problème de taille"
    else:
        C=[[0 for p in cols(B)] for p in lignes(A)];
        for i in lignes(A):
            Ai,Ci = A[i],C[i]
            for j in cols(B):
                tBj = tB[j]
                for k in lignes(tB):
                    Ci[j]+=Ai[k]*tBj[k]
    return C


### Classe d'exception

class ExceptionMatrice(Exception):
    def __init__(self,raison):
        self.raison = raison

    def __str__(self):
        return self.raison

def mprod(A,B):
    if cols(A) != lignes(B):
        raise ExceptionMatrice("Tailles incompatibles")
    else:
        return [[som_liste([A[i][k]*B[k][j] 
                    for k in lignes(B)]) for j in cols(B)] for i in lignes(A)]

    

### matrice identite

def unite(n):
    return [[1 if i==j else 0 for j in range(n)] for i in range(n)]

### puissance d'une matrice

def mpuis(A,n):
    if n == 0: return unite(len(A))
    else: return mprod(A,mpuis(A,n-1))

def mpuis_it_1(A,n):
    pA = unite(len(A))
    for k in range(1,n):
        pA = mprod(pA,A)
    return pA

def mpuis_it(A,n):
    if n == 0: return unite(len(A))
    pA = deepcopy(A)
    k = n - 1
    while k > 0:
        if k%2 == 0:
            pA = mprod(pA,pA)
            k /= 2
        else:
            pA = mprod(A,pA)
            k -= 1
    return pA



### trace

def trace(M):
    if lignes(M) != cols(M):
       raise ExceptionMatrice("La matrice doit être carrée")
    t=0
    for i in lignes(M):
        t += M[i][i]
    return t


### matrice ortho

def est_ortho(M):
    if lignes(M) != cols(M):
        raise ExceptionMatrice("La matrice doit être carrée")
    else:
       for j in cols(M):
       # on parcourt ses vecteurs colonnes
          N = 0
          # la norme d'une colonne est nulle au départ
          for i in lignes(M):
              N += (M[i][j])**2
              if N != 1: return False
              # Si la norme n'est pas 1 on sort
              # chaque colonne doit être de norme 1
          for k in range(j+1,len(cols(M))):
          # on calcule les produits scalaires avec les cols de droite
              P = 0
              for i in lignes(M):
                  P += M[i][j]*M[i][k]
              if P != 0: return False
              # chaque produit scalaire doit être nul
    return True

#### GAUSS

from copy import deepcopy

def comb_ligne(ki,kj,M,i,j):
    """ Lj <- ki*Li + kj*Lj dans la matrice M """
    A = deepcopy(M)
    Lj = [kj*x for x in A[j]]
    Li = [ki*x for x in A[i]]
    A[j] = [Lj[p] + Li[p] for p in cols(M)]
    return A


def mult_lignes(k,M,i):
    """ Li <- k*Li dans la matrice M """
    A = deepcopy(M)
    A[i] = [k * M[i][j] for j in cols(M)]
    return A

def GJ(M):
    """ colle I_n à droite de M """
    U = unite(len(M))
    return [M[i] + U[i] for i in lignes(M)]

def affiche(M):
    for i in lignes(M):
        print(M[i])

def JG(T):
    """ décolle la moitié droite de T"""
    return [T[i][len(T):] for i in lignes(T)]

def swap(M,i,j):
    """ Li <-> Lj """
    S = deepcopy(M)
    S[i] = M[j]
    S[j] = M[i]
    return(S)

def inv_gauss(T):
    """ renvoie M^{-1} par Gauss-Jordan """
    if lignes(T) != cols(T):
      raise ExceptionMatrice("La matrice doit être carrée")
    S = GJ(T)
    for NoColonne in lignes(T):
        NoLigne = NoColonne
        while S[NoLigne][NoColonne] == 0:
            if NoLigne == len(T) - 1:
                raise ExceptionMatrice("Matrice non inversible")
            NoLigne += 1
        S = swap(S,NoLigne,NoColonne)
        pivot = S[NoColonne][NoColonne]
        S = mult_ligne(float(1)/float(pivot),S,NoColonne)
        E = set([k for k in lignes(T)]) - set([NoColonne])
        for k in E:
            S = comb_ligne(-S[k][NoColonne],1,S,NoColonne,k)
    return JG(S)

def tri_gauss2(T):
    S = deepcopy(T)
    mini = min(len(lignes(T)),len(cols(T)))
    for NoColonne in range(mini):
        NoLigne = NoColonne
        while (S[NoLigne][NoColonne]==0) and (NoLigne < len(lignes(T))-1):
                NoLigne += 1
        if S[NoLigne][NoColonne] != 0:
            S = swap(S,NoLigne,NoColonne)
            pivot = S[NoColonne][NoColonne]
            for k in range(NoLigne+1,len(lignes(T))):
                S = comb_ligne(-S[k][NoColonne],pivot,S,NoColonne,k)
    return S


def Im(T):
    S = tri_gauss(transpose(T))
    B = []
    for i in lignes(S):
        if S[i] != [0 for j in cols(S)]: 
            B += [S[i]]
    return B

def rg(T):
    return len(Im(T))

def det_gauss(T):
    if lignes(M) != cols(M):
        raise ExceptionMatrice("La matrice doit être carrée")
    S = T
    i = 1
    for NoColonne in lignes(T):
        NoLigne = NoColonne
        while (S[NoLigne][NoColonne]==0) and (NoLigne<len(T)-1):
            NoLigne += 1
        if S[NoLigne][NoColonne] != 0:
            S = swap(S,NoLigne,NoColonne)
            i *= -1
            pivot = S[NoColonne][NoColonne]
            for k in range(NoLigne+1,len(T)):
                S = comb_ligne(-S[k][NoColonne],pivot,S,NoColonne,k)
                i *= -S[NoColonne][NoColonne]
    d = 1
    for k in lignes(S):
           d *= S[k][k]
    return(d/i)

### HILL

def list2mat(L,p):
    l = len(L)
    complement = -l % p
    LL = L + [0 for k in range(complement+1)]
    # il y a un 0 en trop à cause du ":" de la ligne ci-dessous
    return [LL[p*i:p*(i+1)] for i in range(len(LL)//p)]

def matmod(M,n):
    return [[M[i][j]%n for j in cols(M)] for i in lignes(M) ]

def diag_gauss(T):
    if lignes(T) != cols(T):
        raise ExceptionMatrice("La matrice doit être carrée")
    S = GJ(T)
    for NoColonne in lignes(T):
        NoLigne = NoColonne
        while S[NoLigne][NoColonne] == 0:
            if NoLigne == len(T) - 1:
                raise ExceptionMatrice("Matrice non inversible")
            NoLigne+=1
        S = swap(S,NoLigne,NoColonne)
        pivot = S[NoColonne][NoColonne]
        E = set([k for k in lignes(T)]) - set([NoColonne])
        for k in E:
            S = comb_ligne(-S[k][NoColonne],pivot,S,NoColonne,k)
    return S


def inv_mod_gauss(T):
    if lignes(T) != cols(T):
       raise ExceptionMatrice("La matrice doit etre carree")
    S = GJ(T)
    for NoColonne in lignes(T):
        NoLigne = NoColonne
        while S[NoLigne][NoColonne] == 0:
            if NoLigne == len(T) - 1:
                raise ExceptionMatrice("Matrice non inversible")
            NoLigne += 1
        S = swap(S,NoLigne,NoColonne)
        S = matmod(mult_ligne(inv_mod(S[NoColonne][NoColonne],27),S,NoColonne),27)
        E = set([k for k in lignes(T)]) - set([NoColonne])
        pivot = S[NoColonne][NoColonne]
        for k in E:
            S = comb_ligne(-S[k][NoColonne],pivot,S,NoColonne,k)
    return JG(matmod(S,27))


def mat2list(M):
    L = [];
    for i in lignes(M):
        L += M[i]
    return L


def codage(message,cle):
    L = code(message);
    clair = transpose(matmod(list2mat(L,3),27));
    return decode(mat2list(transpose(matmod(mprod(cle,clair),27))))


def decodage(message_crypte,inv_cle):
    LL = code(message_crypte);
    crypte = matmod(transpose(list2mat(LL,3)),27);
    clarifie = matmod(mprod(inv_cle,crypte),27);
    liste_clarifie = mat2list(transpose(clarifie));
    return decode(liste_clarifie)

### Fibonacci

def u(n):
    if n == 0 or n == 1:
        return 1
    else:
        return u(n-1) + u(n-2)

## pour obtenir U(n)
def mat_fib(n):
    if n == 1: return [[1],[1]]
    else: return mprod([[1,1],[1,0]],mat_fib(n-1))

## pour en extraire un
def fib(n):
    return mat_fib(n)[0][0]






#####################################################
#
#
#      § 2 ENSEMBLES DE NOMBRES
#
#
######################################################""

## partie entière

def partie_entiere(x):
    if x >= 0 and x < 1: return 0
    elif x > 0: return 1 + partie_entiere(x-1)
    else: return -1 + partie_entiere(x+1)

def pe(x):
     k = 0
     if x >= 0:
         while k <= x : k += 1
         return k-1
     else :
         while k > x: k += -1
         return k

### rationnels



     
def pgcd(a,b):
    if b == 0: return a
    else : return pgcd(b,a%b)


class frac:
    def __init__(self,numerateur,denominateur):
        self.num = numerateur
        self.den = denominateur

    def aff(self):
        return str(self.num)+' / '+str(self.den)
    def simp(self):
        if self.den==0:
            raise ZeroDivisionError
        else:
            return frac(self.num//pgcd(self.num,self.den),
                        self.den//pgcd(self.num,self.den))




def som(F1,F2):
    if F1.den == 0 or F2.den == 0:
        raise ZeroDivisionError
    else:
        return frac(F1.num*F2.den + F1.den*F2.num,F1.den*F2.den).simp()

def mul(F1,F2):
    if F1.den == 0 or F2.den == 0:
        raise ZeroDivisionError
    else:
        return frac(F1.num*F2.num,F1.den*F2.den).simp()

def inv(F):
    if F.num == 0 or F.den == 0: raise ZeroDivisionError
    else: return frac(F.den,F.num).simp()

def div(F1,F2):
    return mul(F1,inv(F2))

def quo(a,b):
    if a < b: return 0
    else: return 1 + quo(a-b,b)

def rest(a,b):
    if a < b: return a
    else: return rest(a-b,b)

def fcr(F):
    if F.den == 0 : return []
    else : return [F.num // F.den]+fcr(frac(F.den,F.num % F.den))

def fci(F):
    res = F.num % F.den
    Liste = []
    f = F
    while res > 0:
        Liste += [f.num // f.den]
        f = frac(f.den,res)
        res = f.num % f.den
    return Liste+[f.num // f.den]

def fc2f(L):
    if len(L) == 1: return frac(L[0],1)
    else: return som(frac(L[0],1),inv(fc2f(L[1:])))

def reduite(k,Q):
    if k == 0:
        return frac(Q[0],1)
    elif k == 1:
        return frac(Q[1]*Q[0]+1,Q[1])
    else:
        return frac(Q[k]*(reduite(k-1,Q)).num + 
               (reduite(k-2,Q)).num,Q[k]*(reduite(k-1,Q)).den + 
               (reduite(k-2,Q)).den)

def dif(F1,F2):
    if F1.den == 0 or F2.den == 0:
        raise ZeroDivisionError
    else:
        return frac(F1.num*F2.den - F1.den*F2.num,F1.den*F2.den).simp()



### RÉELS

def heron_rec(a,fo,n):
    if n == 0: return fo
    else:
        return heron_rec(a,div(som(fo,div(frac(a,1),fo)),frac(2,1)),n-1)

def heron_it(a,fo,n):
    app = fo
    for k in range(n):
        app = div(som(app,div(frac(a,1),app)),frac(2,1))


def expor(n):
    if n == 167: return frac(1,1)
    else: return som(frac(1,1),div(expor(n+1),frac(n,1)))

def app_frac(self):
    return self.num/self.den

def diff_newt(x1,y1,y2,y3,y4,pas):
    Dy1 = y2-y1
    Dy2 = y3-y2
    Dy3 = y4-y3
    D2y1 = Dy2-Dy1
    D2y2 = Dy3-Dy2
    D3y1 = D2y2-D2y1
    def P(x):
        return y1+(x-x1)*(Dy1+(x-x1-1)*(D2y1/2+(x-x1-2)*D3y1/6))
    return [[(x1+k*pas),P(x1+k*pas)] for k in range(int(3/pas))]

def dic_log(n):
    dic = {x:log(1+10**(-x)) for x in range(n+1)}
    dic[-1] = log(10)
    return(dic)

def cordic(x,N):
    D = dic_log(N)
    y = D[-1]
    for i in range(N+1):
        a = 1+10**(-i)
        while x*a <= 10:
            x = x*a
            y = y-D[i]
    return y



### COMPLEXES

class cplx:
       def __init__(self,partie_reelle,partie_imaginaire):
              self.r = partie_reelle
              self.i = partie_imaginaire

       def alg(self):
              return str(self.r)+' + '+str(self.i)+'(i)'

def csom(z1,z2):
    return cplx(z1.r+z2.r , z1.i+z2.i)
def cprod(z1,z2):
    return cplx(z1.r*z2.r-z1.i*z2.i , z1.r*z2.i+z1.i*z2.r)

def conj(z):
    return cplx(z.r,-z.i)

def cmod(z):
       return sqrt(z.r*z.r + z.i*z.i)

def cinv(z):
    if z == cplx(0,0): raise ZeroDivisionError
    else:
        return cplx(z.r/(cmod(z))**2 , -z.i/(cmod(z)**2))

def cdiv(z1,z2):
    return cprod(z1,inv(z2))

## Avec matrices


class cplx:
    def __init__(self,partie_reelle,partie_imaginaire):
        self.r = partie_reelle
        self.i = partie_imaginaire
    def mcplx(self):
        return msom(mprod_scal(U,self.r) , mprod_scal(J,self.i))

def alg(Z):
    return str(Z[0][0])+' + '+str(Z[1][0])+'(i)'

def mcsom(Z1,Z2):
    return msom(Z1.mcplx(),Z2.mcplx())

def mcprod(Z1,Z2):
    return mprod(Z1.mcplx(),Z2.mcplx())

def mconj(Z):
    return msom(mprod_scal(U,Z.r) , mprod_scal(J,-Z.i))





#####################################################
#
#
#      § 3 ARITHMÉTIQUE
#
#
######################################################""


## versions récursives :
def quotient(a,b):
    if a < b: return 0
    else: return 1 + quotient(a-b,b)


def reste(a,b):
    if a < b: return a
    else: return reste(a-b,b)

    
# versions itératives:
def quotient(a,b):
    q = 0
    while q*b <= a: q += 1
    return q-1

def reste(a,b):
    q = 0
    while q*b <= a: q += 1
    return a-b*(q-1)


def divise(p,n):
    return n % p == 0


## version iterative
def NbDivNaif(n):
    t=0
    for k in range(1,n+1) :
       if divise(n,k) : t+=1
    return t

def NbDiv(n):
    t = 2 # n et 1 sont des diviseurs de n
    fin = n # le + grand est n
    k = 2
    while k < fin:
       if divise(k,n): 
          fin = n//k # l'autre diviseur est n//k
          if k == fin: t+=1 # cas ou les deux diviseurs sont confondus
          else: t += 2 # sinon, on compte k et n//k 
       k += 1 # si k ne divise pas n, on regarde l'entier suivant
    return t


def NbDivRecTemp(n,k):
    if k*k > n: return 1
    else:
        if (k==1) or (divise(k,n) and (k==n//k)):
            return 1 + NbDivRecTemp(n,k+1)
        else:
            if n%k == 0:
                return 2 + NbDivRecTemp(n,k+1)
            else:
                return NbDivRecTemp(n,k+1)
def NbDivRec(n):
    return NbDivRecTemp(n,1)


def ListeDiv(n):
    L = [1,n]
    fin = n
    k = 2
    while k < fin :
       if n%k == 0 :
          fin = n//k;
          if k == fin : L += [k]
          else : L += [k,n//k]
       k += 1
    return L


def som_list(L):
    S = 0
    for k in range(len(L)):
        S += L[k]
    return S

def parfait(n):
    if som_list(ListeDiv(n)) == 2*n: return True
    else: return False



def Liste_parfait(n):
    L=[]
    for k in range(2,n+1):
       if parfait(k): L += [k]
    return L


def pgcd(a,b):
    if b == 0: return a
    else: return pgcd(b,a%b)

def pgcd2(a,b):
    while b != 0:
       r = a%b
       a = b
       b = r
    return a

def pgcd3(a,b):
    if b == 0: return a
    if a >= b: return pgcd3(b,a-b)
    else: return pgcd3(a,b-a)

def aee(u,v,r,U,V,R):
    if R == 0: return [u,v,r]
    else:
        return aee(U,V,R,u-r//R*U,v-(r//R)*V,r-(r//R)*R)

def AEE(a,b): 
    return aee(1,0,a,0,1,b)


def test_naif(n):
    k = 2
    while k*k <= n:
        if divise(k,n) : return False
        else: k += 1
    return True

def test_local(m,k):
    if m*m > k: return True
    elif divise(k,n): return False
    else: return test_local(m+1,k)

def test_naif(n):
    return test_local(2,n)

def test2(n):
     if n == 2 : return True
     elif divise(2,n): return False
     k = 3
     while k*k <= n:
        if n%k == 0 : return False
        else: k += 2
     return True


## le filtre
def filtre_nul(liste):
    return [liste[i] for i in range(len(liste)) if liste[i]>0]

##le crible
def erato(n):
    P = list(range(n+1))
    P[1] = 0
    j = 2
    while j*j <= n:
       if P[j] >= 1 :
           k = 2
           while k <= n//j:
               P[j*k] = 0
               k += 1
       j += 1
    return filtre_nul(P)


def tete(L):
    return L[0]

def queue(L):
    return L[1:]

def decompotemp(n,liste):
   if n == 1: return []
   elif n % tete(liste) == 0:
      return [tete(liste)] + decompotemp(n//tete(liste),liste)
   else: return decompotemp(n,queue(liste))

def decompo(n):
    return decompotemp(n,erato(n))

def decompo(n):
    L=erato(n) # les nombres premiers
    D=[] # la liste des diviseurs
    N=n # le nombre transforme par divisions successives
    for k in range(len(L)):
        while (N not in L) and (N % L[k] == 0):
                D += [L[k]]
                N = N // L[k]
    return D+[N]

def PI(n):
    return len(erato(n))





def inv_mod(p,n):
    B = bezout(p,n)
    if B[2] == 1:
        return B[0]%n
    else:
        raise ZeroDivisionError



def code(chaine):
    return [ord(chaine[i])-ord(('@')) for i in range(len(chaine))]

def decode(L):
    mot = '';
    for i in range(len(L)):
        mot += chr(L[i]+ord(('@')))
    return mot


def code_affine(message,a,b):
    M = code(message)
    C = [(a*M[i]+b) % 27 for i in range(len(M))]
    return decode(C)

def decode_affine(crypte,a,b):
    C = code(crypte)
    inv_a = inv_mod(a,27)
    D = [( inv_a*(C[i]-b) ) % 27 for i in range(len(C))]
    return decode(D)


def table_vige():
    return [[chr((i+j)%26 + ord('A')) for j in range(26)] for i in range(26)]



def code_vige(message,cle):
    def code(chaine):
        return [ord(chaine[i])-ord(('A')) for i in range(len(chaine))]
    Lm = code(message)
    Lc = code(cle)
    crypte = ''
    for i in range(len(Lm)):
        crypte += chr( (Lm[i]+Lc[i%len(Lc)]) % 26 + ord('A') )
    return crypte


def decode_vige(crypte,cle):
    L = code(crypte)
    Lc = code(cle)
    message = ''
    for i in range(len(L)):
        message += chr( (L[i]-Lc[i%len(Lc)]) % 26 + ord('A') )
    return message


def NbChiffres(n):
    if n<10: return 1
    else: return 1 + NbChiffres(n//10)

def list_ch(n):
    if n < 10: return [n]
    else: return [n%10] + list_ch((n-n%10) // 10)


# somme des cubes des elements d'une liste
def cube_list(L):
    S = 0
    for k in range(len(L)):
        S += L[k]**3
    return S

# listes des "armstrong" inferieurs a n
def armstrong(n):
    A = []
    for k in range(n):
        if cube_list(list_ch(k)) == k:
            A += [k]
    return A

def prod_list(L):
    if len(L) == 1: return L[0]
    else: return L[0]*prod_list(L[1:])


def prod(n):
    return prod_list(list_ch(n))

def prod_list(L):
    P = 1
    for k in range(len(L)):
        P *= L[k]
    return P

def persistance(n):
    if n < 10: return 0
    else:
        return 1 + persistance(prod(n))


def persistance(n):
    N = n
    p = 0
    while N >= 10:
       p += 1
       N = prod(N)
    return p


def test_pers():
    n = 9
    p = 0
    while p < 5:
        n += 1
        p = persistance(n)
    return n

def isbn(code):
    if code[9] == 'X':
        C = [eval(code[i]) for i in range(9)]+[10]
    else:
        C = [eval(code[i]) for i in range(10)]
    s = 0
    for i in range(10):
        s += (i+1)*C[9-i]
    return s%11 == 0

def cle_isbn(code_sans_cle):
    C = [eval(code_sans_cle[i]) for i in range(9)]
    s = 0
    for i in range(1,10):
        s += (i+1)*C[9-i]
    cle = -s % 11
    if cle == 10: return 'X'
    else: return cle

def base(n,b):
    r = n%b
    q = n//b
    R = [r]
    while q>0:
        r = q%b
        q = q//b
        R = [r]+R
    return R

def base(n,b):
    if n < b: return [n]
    else: return base(n//b,b) + [n%b]

def puissance_rapide(a,n):
    C = base(n,2)
    temp = a
    for j in range(1,len(C)):
        temp *= temp
        if C[j] == 1: temp *= a
    return(temp)


def pmod(a,m,p):
    C = base(m,2)
    temp = a%p
    for j in range(1,len(C)):
        temp *= temp%p
        if C[j] == 1: temp *= a%p
    return(temp%p)


def liste_test_fermat(a,n):
    E = erato(n)
    for p in E:
        print(p,pmod(a,p-1,p)==1)


def test_fermat(a,n):
    E = erato(n)
    test = True
    for p in E:
        if a%p != 0:
            test = test and (pmod(a,p-1,p)==1)
    return test

def pseudo(a,n):
    E = erato(n)
    L = []
    for i in range(3,n,2):
        if not(i in E) and (pmod(a,i-1,i)==1):
            L += [i]
    return L




def bibine(n):
    bi = base(n,2)
    while len(bi) % 4 != 0:
        bi = [0]+bi
    bibi = ""
    for k in range(len(bi),0,-4):
        if bi[k-2:k] == [0,0]: bibi = 'O'+bibi
        elif bi[k-2:k] == [0,1]: bibi = 'A'+bibi
        elif bi[k-2:k] == [1,0]: bibi = 'E'+bibi
        else: bibi = 'I'+bibi
        if bi[k-4:k-2] == [0,0]: bibi = 'H'+bibi
        elif bi[k-4:k-2] == [0,1]: bibi = 'B'+bibi
        elif bi[k-4:k-2] == [1,0]: bibi = 'K'+bibi
        else: bibi = 'D'+bibi
    return bibi



def enibib(L):
    decod = {bibine(k):k for k in range(16)}
    N = 0
    for k in range(0,len(L)-2,2):
        N += decod[L[k:k+2]] * 16**((len(L)-k+1)//2-1)
    return N










#####################################################
#
#
#      § 4 PI
#
#
######################################################""

def cues():
    h = 1/4
    r = sqrt(2)/4
    compteur = 0
    while not(pseudo_egal_float(1/h,1/r)):
        h = (r+h)/2
        r = sqrt(r*h)
        compteur += 1
    return 1/r,1/h,1/h-1/r,compteur

def int_rec_droite(f,a,b,N):
    S = 0
    t = a
    dt = (b-a)/N
    while t < b:
        S += f(t)*dt
        t += dt
    return S


def int_rec_gauche(f,a,b,N):
    S = 0
    t = a
    dt = (b-a)/N
    while t+dt < b:
        S += f(t+dt)*dt
        t += dt
    return S


def int_trap(f,a,b,N):
    S = 0
    t = a
    dt = (b-a)/N
    while t+dt <= b:
        S += (f(t)+f(t+dt))*dt/2
        t += dt
    return S


def int_trap(f,a,b,N):
    S = 0
    dt = (b-a)/N
    for k in range(N):
        S+=(f(a+k*dt)+f(a+(k+1)*dt))*dt/2
    return S


def int_simps(f,a,b,N):
    S = 0
    t = a
    dt = (b-a)/N
    while t+dt <= b:
        S += (f(t)+4*f(t+dt/2)+f(t+dt))*dt/6
        t += dt
    return S


def mini_greg(a,d):
    getcontext().prec = d+2
    n = 1
    ad = Decimal('1')/Decimal(str(a))
    D = Decimal('1e-'+str(d))
    nd = ad**3
    dd = Decimal('3')
    fd = nd/ad
    while fd > D :
        nd *= ad*ad
        dd += 2
        fd = nd/ad
        n += 1
    return n


def greg(a,N):
    ad = Decimal('1')/Decimal(str(a))
    nd = ad
    dd = Decimal('1')
    s = nd/dd
    for k in range(1,N):
        nd *= Decimal('-1')*ad*ad
        dd += Decimal('2')
        s += nd/dd
    return s

def pi_machin(d):
    getcontext().prec = d+2
    rap = 2*log(5)/log(10)
    N = int(d/rap)+1
    return 4*(4*greg(5,N)-greg(239,N))


def dissec(n):
    N = n*n+1
    L = [[1,N],[-1,-N]]
    if N%2 == 0:
        L += [[2,N//2],[-2,-N//2]]
    else:
        for k in range(3,N//2,2):
            if N%k == 0:
                L += [[k,N//k],[-k,-N//k]]
    for j in range(len(L)):
        if not(-n in L[j]):
            print('Arctan 1/',n,'= Arctan 1/',L[j][0]+n,' + Arctan 1/',L[j][1]+n,end = '\n')



def som(L):
    S = 0
    for j in range(len(L)):
        S += L[j]
    return S

def greg(p,C,A):
    rap = 2*log(A[0])/log(10)
    N = int(p/rap)+1
    s = (-1)**N
    getcontext().prec = p+2
    cx = [s*Decimal(str(C[j])) for j in range(len(C))]
    ax = [Decimal(str(A[j])) for j in range(len(C))]
    X = [cx[j]/ax[j]**(2*N+1) for j in range(len(C))]
    kx = [ax[j]*ax[j] for j in range(len(C))]
    S = som(X)/(2*N+1)
    for k in range(2*N-1,0,-2):
        X = [-X[j]*kx[j] for j in range(len(C))]
        S += som(X)/Decimal(str(k))
    return S


def greg(p,C,A):
    t1 = time()
    rap = 2*log(A[0])/log(10)
    N = int(p/rap)+1
    s = (-1)**N
    getcontext().prec = p+2
    cx = [s*Decimal(str(C[j])) for j in range(len(C))]
    ax = [Decimal(str(A[j])) for j in range(len(C))]
    X = [cx[j]/ax[j]**(2*N+1) for j in range(len(C))]
    kx = [ax[j]*ax[j] for j in range(len(C))]
    S = som(X)/(2*N+1)
    for k in range(2*N-1,0,-2):
        X = [-X[j]*kx[j] for j in range(len(C))]
        S += som(X)/Decimal(str(k))
    t2 = time()
    return t2-t1


def brent(n):
    d = Decimal('2')
    a = Decimal('1')
    u = Decimal('0.5')
    s = u
    b = s*d.sqrt()
    c,p,k=a*a-b*b,2*a*a/s,1
    for j in range(n):
        a,b = (a+b)*u,(a*b).sqrt()
        c,k = a*a-b*b,2*k
        s = s-c*k
    return d*a*a/s





#####################################################
#
#
#      § 5 Probas
#
#
######################################################""


def piece(n):
    T = []
    for k in range(n):
        X = 0
        for j in range(3):
            X += randint(0,1)
        T += [X]
    return([(T.count(i)/float(n)) for i in range(0,4) ])


def boules(n):
    T = []
    for k in range(n):
        X = 0
        if randint(1,11) <= 7: X += 1
        if randint(1,7) <= 5: X += 1
        if randint(1,9) <= 6: X += 1
        T+=[X]
    return([(T.count(j)/float(n)) for j in range(0,4) ])


def boule(n,nb):
    X=0
    for k in range(n):
        B = 0
        for j in range(1,5):
            if randint(1,10) <= 7: B += 1
        if B == nb: X += 1
    return X/n


def toscane(n):
    T = n*[0]
    for k in range(n):
        T[k] = randint(1,6) + randint(1,6) + randint(1,6)
    return 'Obtenir 9 :' + str(T.count(9)*100.0/n) + '%, Obtenir 10 :' + str(T.count(10)*100.0/n) + '%'


def piece_max(n):
    S = n*[1]
    for k in range(n):
        s = []
        P = [randint(0,1) for z in range(10)]
        p = 0
        while p < 9:
            j = p+1
            while (P[j] == P[p] and j < 9):
                j += 1
            s += [j-p]
            p = j
        s.sort(reverse=True)
        S[k] = s[0]
    m = (sum(S)/n)
    return 'Sur '+str(n)+'  séries de 10 lancers, la moyenne du nombre maximal de résultats consécutifs égaux est '+str(m)



def jeutortue(n):
    [L,T] = [0,0]
    for simul in range(n):
        face = randint(1,6)
        lancer = 1
        while face < 6 and lancer < 6:
            face=randint(1,6)
            lancer += 1
        if face == 6 : L += 1
        else : T += 1
    return 'Tortue :'+str(T*100.0/n)+'% Lièvre : '+str(L*100.0/n)+'%'



def monty(n):
    gagne_sans_changer = 0
    gagne_en_changeant = 0
    for j in range(n):
        voiture = randint(0,2)
        choix = randint(0,2)
        if choix == voiture:
            ouverte = (voiture+1+randint(0,1)) % 3
        else:
            ouverte = 0+1+2-choix-voiture
        changement = 0+1+2-choix-ouverte
        if choix == voiture:
            gagne_sans_changer += 1
        if changement == voiture:
            gagne_en_changeant += 1
    return 'Gagne en changeant : '+str(100.*gagne_en_changeant/n)+'%   Gagne sans changer : '+ str(100.*gagne_sans_changer/n)+'%'






#####################################################
#
#
#      § 6 Graphes
#
#
######################################################""

def compose(R,S):
    """ R suivi de S """
    C = {} # Le dictionnaire de la composée est vide au départ
    for u in R: # pour chaque sommet u de E
        C[u] = set([]) # u n'est encore en relation avec personne par C
        for v in R[u]: # pour chaque v tel que uRv
            C[u] |= S[v] # uCw <-> vSw
    return C


def ens(R):
    return set(list(R))

def id(E):
    D = {}
    for u in E: # chaque sommet u vérifie uRu
        D[u] = set([u])
    return D


def iter_compo(R,n):
    """ R^n version récursive """
    if n == 0:
        return id(ens(R)) # R^0 = id
    else: # sinon R^n = R.R^(n-1)
        return compose(R,iter_compo(R,n-1))


def est_reflexive(R):
    for u in R:
        if not(u in R[u]): # dès qu'on trouve un contre-exemple, on sort
            return False
    return True # si on n'est pas déjà sorti c'est que la relation est réflexive

def est_symetrique(R):
    for u in R:
        if u not in [R[v] for v in R[u]]:
            return False
    return True


def est_antisymetrique(R):
    for u in R:
        if u in [R[v] for v in R[u] if v != u]:
            return False
    return True


def est_transitive(R):
    R2 = compose(R,R)
    for u in R2:
        if not(R2[u] <= R[u]):
            return False
    return True


def r_plus(R):
    Rp = deepcopy(R)
    for k in range(1,len(Rp)+1):
        Rk = iter_compo(R,k)
        for u in Rp:
            Rp[u] |= Rk[u]
    return Rp


from copy import deepcopy

def bellman_ford(graphe, source):
    """ Programme direct sans les chemins """
    pi = {}
    for s in graphe:
        pi[s] = float('inf')
    pi[source] = 0
    pi0 = {}
    k = 1
    while pi != pi0 :
        pi0 = deepcopy(pi)
        for u in graphe:
            for v in graphe[u]:
                if pi[v] > pi[u] + graphe[u][v]:
                    pi[v]  = pi[u] + graphe[u][v]
        k+=1
    return pi


def plus_court(graphe,debut,fin,visites,pi,chaine,espion):
    """trouve la plus courte chaine entre debut et fin avec l'algo de Dijkstra
       visites et chaine sont des listes et pi un dictionnaire
       version récursive
    """
    # si on arrive à la fin, on affiche la distance et la chaine en y rajoutant fin
    if debut==fin:
       chaine.append(fin)
       return pi[fin], chaine, espion
    # si c'est la première visite, on met pi(debut) à 0
    if  len(visites) == 0 : pi[debut]=0
    # on commence à tester les voisins
    for voisin in graphe[debut]:
        if voisin not in visites:
            dist_voisin = pi.get(voisin,float('inf'))
            candidat_dist = pi[debut] + graphe[debut][voisin]
            if candidat_dist < dist_voisin:
                pi[voisin] = candidat_dist
                if debut not in chaine:
                    chaine.append(debut)
    # on a regardé tous les voisins : le noeud entier est visité
    visites.append(debut)
    espion+=1
    # on cherche le sommet non visité le plus proche
    non_visites = dict((s, pi.get(s,float('inf'))) for s in graphe if s not in visites)
    noeud_plus_proche = min(non_visites, key=non_visites.get)
    # on applique récursivement en prenant comme début le plus proche
    return plus_court(graphe,noeud_plus_proche,fin,visites,pi,chaine,espion)


def dij_rec(graphe,debut,fin):
    return plus_court(graphe,debut,fin,[],{},[],1)


def colo(g):
    n = len(g) # nb de sommets
    couleurs_dispo = set(range(1,n+1)) # ensemble des couleurs disponibles
    g_color = {} # le dictionnaire des sommets colorés avec leurs couleurs
    for u in g: # pour chaque sommet du graphe
        couleurs_interdites  = set([g_color[v] for v in g[u] if v in g_color])
        # on interdit les couleurs des adjacents
        g_color[u] = min(couleurs_dispo - couleurs_interdites)
        # on choisit celle qui porte le plus petit numéro
    return g_color,max(g_color.values())


from random import shuffle

def colo_rand(g):
    n = len(g) # nb de sommets
    dic_col = {} # dictionnaire des meilleures colorations indexé par le nb chromatique provisoire
    liste_sommets = list(g)
    for k in range(1000):
        shuffle(liste_sommets,random=random.random) # les sommets mélangés en place
        couleurs_dispo = set(range(1,n+1)) # ens des couleurs disponibles
        g_color = {} # le dic des sommets colorés avec leurs couleurs
        for u in liste_sommets: # pour chaque sommet du graphe ordre aléatoire
            couleurs_interdites = set([g_color[v] for v in g[u] if v in g_color])
            g_color[u] = min(couleurs_dispo - couleurs_interdites)
        m = max(g_color.values())
        dic_col[m] = g_color # on associe le dictionnaire de coloration à chi_provisoire
    mi = min(dic_col.keys()) # le plus petit nb chromatique provisoire
    return 'chi(g) <= '+str(mi),dic_col[mi]  # on affiche une plus petite coloration possible




